#
# Higher level plotting functions (i.e. generating multiple plots)
# lower level plot functions are in banksy_utils.plotting
# Refactored/edited by Yifei
#

import numpy as np
from numpy.random import randint
import pandas as pd

import os
import gc
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import seaborn as sns
from anndata import AnnData
import anndata
from banksy_utils.plotting import plot_edge_histogram, plot_graph_weights
from banksy_utils.cluster_utils import pad_clusters, get_DEgenes, get_metagene_difference

from banksy_utils.slideseq_ref_data import dropviz_dict, markergenes_dict

import scanpy as sc

from scipy.stats import pearsonr, pointbiserialr
from itertools import cycle
from typing import Union, Tuple, List
import scipy.sparse as sparse


def plot_qc_hist(
        adata: anndata.AnnData,
        total_counts_cutoff: int,
        n_genes_high_cutoff: int,
        n_genes_low_cutoff: int,
        **kwargs):
    '''
    Plots a 1-by-4 figure of histograms for
    1: total_counts (unfiltered), 2: total_counts (filtered), 3: n_genes (unfiltered), 4: n_genes (filtered)

    Args:
        Anndata: AnnData Object containing 'total_counts' and 'n_genes_by_counts' 
        total_counts_cutoff: Lower threshold for total counts, if 
        n_genes_high_thrsehold: Upper threshold of gene count

    Optional Args:
        figsize; default = (15,4): size of the figures
        bins1-4; default = 'auto': the number of bins for plotting the histograms (e.g., bin1=80 sets the 80 bins for the first histogram)
    '''
    # Edit bin_size_for_plts as a option, remove bins1,2
    options = {'figsize': (15, 4),
               'bin_options': ['auto', 'auto', 'auto', 'auto']}

    options.update(kwargs)
    bin_sizes = options['bin_options']

    fig, axs = plt.subplots(1, 4, figsize=(15, 4))

    sns.histplot(adata.obs["total_counts"], kde=False, bins=bin_sizes[0], ax=axs[0]).set(
        title='Total Counts (unbounded)')

    sns.histplot(adata.obs["total_counts"][adata.obs["total_counts"] < total_counts_cutoff],
                 kde=False, bins=bin_sizes[1], ax=axs[1]).set(title=f'Total Counts (bounded)')

    sns.histplot(adata.obs["n_genes_by_counts"], kde=False, bins=bin_sizes[2], ax=axs[2]).set(
        title='n-genes by counts (unbounded)')

    sns.histplot(adata.obs["n_genes_by_counts"][
                     (adata.obs["n_genes_by_counts"] < n_genes_high_cutoff) & (
                             adata.obs["n_genes_by_counts"] > n_genes_low_cutoff)
                     ], kde=False, bins=bin_sizes[3], ax=axs[3]).set(title='n-genes by counts (bounded)')

    fig.tight_layout()
    gc.collect()


def plot_cell_positions(adata: anndata.AnnData,
                        raw_x: pd.Series,
                        raw_y: pd.Series,
                        coord_keys: Tuple[str],
                        **kwargs) -> None:
    """
    Plots the position of cells in the dataset

    Args: 
        adata - Anndata containing information about the cell
        raw_x - Positions of the cells
        raw_y

    Optional Args:
        add_cricle: adding a circle patch for AFT data

    Returns scatter plot of the cell positions 
    """
    options = {
        'fig_size': (8, 8),
        's': 0.2,
        'c1': 'red',
        'c2': 'slateblue',
        'label1': 'Original cells',
        'label2': 'Remaining cells',
        'add_circle': False,
        'puck_center': (3330, 3180),
        'puck_radius': 2550
    }

    options.update(kwargs)
    print(options)

    fig, ax = plt.subplots(figsize=options['fig_size'])
    ax.scatter(raw_x, raw_y, s=options['s'],
               c=options['c1'], label=options['label1'])
    ax.scatter(adata.obs[coord_keys[0]], adata.obs[coord_keys[1]],
               s=options['s'], c=options['c2'], label=options['label2'])
    ax.set_title("Cell positions")
    if options['add_circle']:
        circle = plt.Circle(
            options['puck_center'], options['puck_radius'], color='g', fill=False)
        ax.add_patch(circle)

    ax.axis("equal")
    gc.collect()


def plot_edge_histograms(distance_graph: sparse.csr_matrix,
                         weights_graph: sparse.csr_matrix,
                         decay_type: str,
                         m: int,
                         **kwargs) -> None:
    '''An utility function to plot subplots of each histograms
    using the plot_edge_histogram function for each distance and weights'''

    options = {'rows': 1, 'cols': 2, 'figsize': (10, 5)}
    options.update(kwargs)

    # Plots the histograms of distance and weights each m
    fig, ax = plt.subplots(
        nrows=options['rows'], ncols=options['cols'], figsize=options['figsize'])
    plot_edge_histogram(distance_graph, ax[0], title="distances between cells")
    plot_edge_histogram(weights_graph, ax[1], title="weights between cells")
    fig.suptitle(f"nbr decay type = {decay_type}, m = {m}")
    fig.tight_layout()


def plot_weights(
        adata: anndata.AnnData,
        processing_dict: dict,
        nbr_weight_decay: str,
        max_m: int,
        fig_title: str,
        coord_keys: Tuple[str],
        theta_graph: sparse.csr_matrix = None,
        **kwargs):
    '''Plots the weighted graphs using the plotting function
    Optional Args: theta_graph, max_weight, markersize, figsize
    '''

    options = {'max_weight': 0.5, 'markersize': 1.5, 'figsize': (15, 8)}
    options.update(kwargs)

    for m in range(max_m + 1):
        # Plots the weighted graph for each iterations of m
        plot_graph_weights(
            adata.obsm[coord_keys[2]],
            processing_dict[nbr_weight_decay]["weights"][m],
            theta_graph=theta_graph,
            max_weight=options["max_weight"],
            markersize=options["markersize"],
            figsize=options["figsize"],
            title=f"{fig_title}, m = {m}"
        )

        ax = plt.gca()
        ax.axis("equal")


def plot_theta_graph(adata: anndata.AnnData,
                     theta_graph: sparse.csr_matrix,
                     coord_keys: Tuple[str],
                     **kwargs):
    '''Shows the theta graph of the neigbourhood of a random cell'''

    options = {'alpha': 0.4, 'figsize': (6, 6)}
    options.update(kwargs)

    key = coord_keys[2]
    for cell in randint(1, adata.shape[0], 1):
        fig, ax = plt.subplots(figsize=options['figsize'])
        ax.plot(adata.obsm[key][cell, 0], adata.obsm[key]
        [cell, 1], "bo", alpha=options['alpha'])
        ptr_start = theta_graph.indptr[cell]
        ptr_end = theta_graph.indptr[cell + 1]
        for nbr, angle in zip(theta_graph.indices[ptr_start:ptr_end], theta_graph.data[ptr_start:ptr_end]):
            ax.plot(adata.obsm[key][nbr, 0], adata.obsm[key]
            [nbr, 1], "ro", alpha=options['alpha'])
            ax.annotate(f"{angle / np.pi * 180:0.1f}",
                        (adata.obsm[key][nbr, 0], adata.obsm[key][nbr, 1]))
            ax.axis("equal")

        ax.set_title(f"Azimuthal Angles in a cell's neighborhood")

    gc.collect()


def plot_DE_genes(adata_spatial: anndata.AnnData,
                  adata_nonspatial: anndata.AnnData,
                  plot_heatmap: bool = False,
                  plot_dot_plot: bool = False,
                  ) -> None:
    '''
    Function to plot the Differentially-Expressed (DE) genes per group

    Returns adata_nonspatial, adata_spatial subsequent plotting if needed
    '''

    for adata in [adata_spatial, adata_nonspatial]:

        # Plot Differentially Expressed (DE) genes per group
        # --------------------------------------------------
        sc.tl.rank_genes_groups(
            adata, groupby='cell type', method='wilcoxon', verbose=True
        )
        sc.tl.dendrogram(adata, groupby='cell type')

        if plot_heatmap:
            sc.pl.rank_genes_groups_heatmap(adata, n_genes=6,
                                            vmin=-3, vmax=3, cmap='bwr',
                                            swap_axes=True)
        if plot_dot_plot:
            sc.pl.rank_genes_groups_dotplot(adata, n_genes=5)

    gc.collect()


def _adata_filter_self(adata: anndata.AnnData,
                       is_nbr_colname: str = "is_nbr",
                       run_rank_genes_groups: bool = True,
                       cell_type_colname: str = "cell type refined", ) -> anndata.AnnData:
    """
    Filter AnnData object to remove all neighbour features (mean/AGF)
    leaving only self-expresssion features
    Assumes AnnData object has .var column of type bool that defines if each feature is self or neighbour
    """

    adata_self = adata[:, adata.var[is_nbr_colname] == False].copy()

    if run_rank_genes_groups:
        sc.tl.rank_genes_groups(
            adata_self, groupby=cell_type_colname, method='wilcoxon', verbose=True
        )
        sc.tl.dendrogram(adata_self, groupby=cell_type_colname)

    return adata_self


def plot_DE_genes_refined(adata_spatial: anndata.AnnData,
                          plot_heatmap: bool = True,
                          plot_dotplot: bool = False,
                          n_genes: int = 5,
                          cell_type_colname: str = "cell type refined",
                          save_fig: bool = False,
                          file_path: str = "",
                          save_name: str = "",
                          **kwargs) -> anndata.AnnData:
    '''
    Function to plot Differentially Expressed (DE) genes 
    by self-expression for the refined anndata

    Args:
        adata_spatial: this should be the BANKSY matrix

    Optional args:
        save_fig (default = False)
        kwargs: optional specifications for saving figure
    '''

    options = {
        'fig_size': (12, 5),
        'save_format': 'eps',
        'dpi': 1200
    }
    options.update(kwargs)

    # filter AnnData object for only self-expresssion features
    adata_self = _adata_filter_self(adata_spatial,
                                    run_rank_genes_groups=True,
                                    cell_type_colname=cell_type_colname, )

    if plot_heatmap:

        if save_fig:
            save_path = os.path.join(file_path, save_name + options['save_format'])
        else:
            save_path = None

        sc.pl.rank_genes_groups_heatmap(adata_self, n_genes=n_genes,
                                        vmin=-3, vmax=3, cmap='bwr', swap_axes=True,
                                        save=save_path)

    if plot_dotplot:
        fig2, ax = plt.subplots(figsize=options['fig_size'])

        sc.pl.rank_genes_groups_dotplot(adata_self, n_genes=n_genes, groups=None, ax=ax)

        if save_fig:
            fig2.savefig(os.path.join(file_path, save_name + "dotplot." + options['save_format']),
                         format=options['save_format'], dpi=options['dpi'])

    return adata_self


def plot_connection_grid(
        weights: sparse.csr_matrix,
        adata_nonspatial: anndata.AnnData,
        adata_spatial: anndata.AnnData,
        cell_types: list,  # list of clusters/cell types to compare
        cell_types_colname: str = "cell type refined",
        cmap: str = "viridis",
        save_fig: bool = False,
        file_path: str = "",
        save_name: str = "",
        **kwargs) -> None:
    # Changed weights to be argument instead of processing_dict;
    # weights = processing_dict["scaled_gaussian"]["weights"][0].copy()
    '''
    Plots connection grid showing intermingling extent between spatially
    distinct layers
    '''

    options = {
        'fig_size': (12, 5),
        'save_format': 'eps',
        'dpi': 1200
    }
    options.update(kwargs)

    print(
        f"All refined cell types: {adata_spatial.obs[cell_types_colname].cat.categories}")

    fig, axs = plt.subplots(nrows=1, ncols=2,
                            figsize=(len(cell_types) * 2.4 + 3, len(cell_types) * 1.2))

    connection_grids = []

    max_nondiag = 0.0

    for adata in [adata_nonspatial, adata_spatial]:

        one_hots = []
        for cell_type in cell_types:
            one_hots.append(
                np.array(adata.obs[cell_types_colname] == cell_type))

        # one hot labels
        # --------------
        one_hot_array = np.stack(one_hots, axis=0)
        # print(one_hot_array.shape, one_hot_array.dtype)

        # adjacency matrix
        # ----------------
        # use the m=0 weights matrix (default 15 nearest nbrs)
        weights.data = np.ones_like(weights.data)
        print(weights.data, weights.shape)

        # Unnormalized summed connections
        # -------------------------------
        connection_grid = one_hot_array @ weights @ one_hot_array.T

        # Normalize
        # ---------
        connection_grid /= np.diag(connection_grid)[:, np.newaxis]

        connection_grids.append(connection_grid)

        # find max non-diagonal by zeroing out diagonal values (all diagonals = 1)
        connections_zerod = connection_grid.copy()
        np.fill_diagonal(connections_zerod, 0)
        max_nondiag = max(max_nondiag, np.amax(connections_zerod))

    for n, connection_grid in enumerate(connection_grids):
        heat_map = sns.heatmap(connection_grid,
                               linewidth=1,
                               vmax=1,
                               cmap=cmap, annot=True, fmt="0.2f",
                               yticklabels=cell_types, xticklabels=cell_types,
                               square=True, ax=axs[n])
    gc.collect()
    fig.tight_layout()

    if save_fig:
        save_location = os.path.join(file_path, save_name)
        fig.savefig(save_location, format=options['save_format'], dpi=options['dpi'])


def plot_clusters(adata_spatial: anndata.AnnData,
                  adata_nonspatial: anndata.AnnData,
                  colors_dict: dict,
                  clusters_colname: str = "cell type",
                  save_fig: bool = False,
                  file_path: str = "",
                  save_name: str = "",
                  **kwargs) -> None:
    '''
    Plot spatial maps of clusters for spatial and nonspatial, side by side
    (or any 2 clustering outputs you want to compare)
    '''

    options = {
        'fig_size': (8, 6),
        'save_format': 'eps',
        'dpi': 1200,
        'key': 'coord_xy',
        'title': ['Spatial', "Nonspatial"]
    }
    options.update(kwargs)

    coords_key = options['key']

    fig, axs = plt.subplots(
        nrows=1, ncols=2, figsize=options['fig_size'], constrained_layout=True
    )

    for n, adata in enumerate([adata_spatial, adata_nonspatial]):

        ax_loc = axs[n]

        # umap_name = f"reduced_pc_{pca_dims[0]:2d}_umap"

        for layer in adata.obs[clusters_colname].cat.categories:

            if layer in colors_dict:
                c = colors_dict[layer]
                locations = adata.obsm[coords_key][adata.obs[clusters_colname] == layer, :]
                # umap_embedding = adata.obsm[umap_name][adata.obs[clusters_colname] == layer, :]
                loc_plot = ax_loc.scatter(
                    locations[:, 0], locations[:, 1], c=c, s=.2, alpha=1.0
                )

        ax_loc.set_aspect('equal', 'datalim')
        ax_loc.axis("off")
        ax_loc.set_title(options['title'][n])

    if save_fig:
        save_location = os.path.join(file_path, save_name)
        fig.savefig(save_location, format=options['save_format'], dpi=options['dpi'])


def _plot_embedding(embedding: str,  # key for obsm entry in an AnnData object
                    ax: plt.Axes,
                    adata: AnnData,
                    cluster_subset: list,
                    spot_size: float,
                    colours_dict: dict,
                    cluster_colname: str = "cell type",
                    ) -> None:
    """
    Plot an embedding in spatial or expression umap axes

    The embedding should be stored as an obsm (i.e. multidimensional observation data array)
    entry in the AnnData object provided as adata.
    Only first 2 dimenstions of the embedding will be plotted

    plots clusters not in 'cluster_subset' in gray
    """

    # plot other clusters in gray first
    # (to make sure spots are below those of desired clusters in case of overlap)

    for cluster in adata.obs[cluster_colname].cat.categories:

        if cluster not in cluster_subset:
            c = "whitesmoke"
            locations = adata.obsm[embedding][adata.obs[cluster_colname] == cluster, :]
            scatterplot = ax.scatter(locations[:, 0], locations[:, 1], c=c,
                                     s=spot_size, alpha=1.0)

    # then plot the clusters in 'cluster subset'

    for cluster in adata.obs[cluster_colname].cat.categories:

        if cluster in cluster_subset:
            c = colours_dict[cluster]
            locations = adata.obsm[embedding][adata.obs[cluster_colname] == cluster, :]
            scatterplot = ax.scatter(locations[:, 0], locations[:, 1], c=c,
                                     s=spot_size, alpha=1.0)


def _plot_continuous(x, y, c, ax,
                     spot_size: float = 0.3,
                     vmax: float = None,
                     cmap: str = "Blues",
                     title: str = None,
                     plot_cbar: bool = True,
                     ) -> None:
    """
    Plot continous-valued location plot (e.g. markergene, RCTD weights)
    """

    scatterplot = ax.scatter(x, y,
                             c=c, s=spot_size,
                             alpha=1.0, cmap=cmap,
                             vmax=vmax, vmin=0,
                             )

    ax.set_aspect('equal', 'datalim')

    if title is not None:
        ax.set_title(title, fontsize=8, fontweight="bold", )
    # ax.set_ylim(ax_locs.get_ylim()[::-1])
    ax.axis("off")

    if plot_cbar:
        cax = ax.inset_axes([0.8, 0.0, 0.16, 0.04])
        # cax.axis("off")
        ax.figure.colorbar(scatterplot, cax=cax, orientation='horizontal')


def plot_markergene_sets(metagene_df: pd.DataFrame,
                         rctd_coord: pd.DataFrame,
                         rctd_weights: pd.DataFrame,
                         coord_keys: list,  # keys for x and y columns
                         reference_dict: dict = dropviz_dict,  # mapping celltype names to ref names
                         markergenes_dict: dict = markergenes_dict,  # lists of markergenes for each celltype
                         plot_cbar: bool = False,
                         save_fig: bool = False,
                         file_path: str = "",
                         save_name: str = "",
                         **kwargs) -> None:
    '''
    Plots all the marker gene sets for scRNA-seq clusters 

    defaults to labels and DE genes from dropviz as a reference.
    '''
    options = {
        'subplot_size': (3, 3),
        'save_format': 'png',
        'dpi': 1200
    }
    options.update(kwargs)

    subplot_size = options['subplot_size']
    num_cols = 2
    num_rows = len(reference_dict) // num_cols + 1
    figsize_x = (subplot_size[0] * num_cols) * 2.1
    figsize_y = subplot_size[1] * num_rows

    fig = plt.figure(figsize=(figsize_x, figsize_y), constrained_layout=True)
    grid = fig.add_gridspec(ncols=num_cols * 2, nrows=num_rows)

    for n, layer in enumerate(reference_dict):

        # plot the RCTD weights

        ax = fig.add_subplot(grid[n // num_cols, 2 * (n % num_cols)])

        clust_num = reference_dict[layer] - 1
        weights = np.array(rctd_weights[:, clust_num]).flatten()

        _plot_continuous(rctd_coord["x"], rctd_coord["y"], weights,
                         ax, vmax=1, title=layer + "\nRCTD weights",
                         spot_size=0.2, plot_cbar=plot_cbar,
                         )

        if layer in markergenes_dict:

            # if available, plot the meta-gene weights for the marker genes

            ax = fig.add_subplot(grid[n // num_cols, 2 * (n % num_cols) + 1])

            c = metagene_df[layer]
            if isinstance(c, pd.Series) and c.isnull().any():
                c.fillna(0, inplace=True)

            _plot_continuous(metagene_df[coord_keys[0]], metagene_df[coord_keys[1]], c,
                             ax, vmax=1, title=layer + "\nmarker genes",
                             spot_size=0.2, plot_cbar=plot_cbar,
                             )
    if save_fig:
        save_location = os.path.join(file_path, save_name)
        fig.savefig(save_location, format=options['save_format'], dpi=options['dpi'])


def plot_cluster_subset(cluster_subset: list,
                        colours_dict: dict,
                        adata: anndata.AnnData,
                        ax: plt.Axes,
                        title: str = None,
                        umap_name: str = "umap",
                        spot_size: float = 0.4,
                        plot_umap: bool = False,
                        ) -> None:
    """
    Plot a subset of clusters using specified colours from colours_dict
    """

    # plot spatial locations
    _plot_embedding("coord_xy", ax, adata, cluster_subset, spot_size, colours_dict)
    ax.set_aspect('equal', 'datalim')

    if title is not None:
        ax.set_title(title, fontsize=20, fontweight="bold", )
    # ax.set_ylim(ax_locs.get_ylim()[::-1])

    ax.axis("off")

    # plot  UMAP as inset
    if plot_umap:
        axins = ax.inset_axes([0.9, 0.0, 0.25, 0.25])
        _plot_embedding(umap_name, axins, adata, cluster_subset, spot_size, colours_dict)
        axins.set_aspect('equal', 'datalim')
        # axins.axis("off")
        axins.set_ylabel("UMAP2", fontsize=8, fontweight="bold", )
        axins.set_xlabel("UMAP1", fontsize=8, fontweight="bold", )
        axins.spines['right'].set_visible(False)
        axins.spines['top'].set_visible(False)
        axins.xaxis.set_ticks([])
        axins.yaxis.set_ticks([])


def plot_weights_comparsion(adata_spatial: anndata.AnnData,
                            adata_nonspatial: anndata.AnnData,
                            adata_allgenes: anndata.AnnData,  # unfiltered Anndata object to plot marker gene expr
                            cluster_keys: List[list],  # keys for spatial, nonspatial, RCTD, markergenes
                            colours_dict: dict,
                            rctd_coord: pd.DataFrame = None,
                            rctd_weights: pd.DataFrame = None,
                            reference_dict: dict = dropviz_dict,  # mapping celltype names to ref names
                            markergenes_dict: dict = markergenes_dict,  # lists of markergenes for each celltype
                            save_fig: bool = False,
                            file_path: Union[str, None] = None,
                            save_png_name: Union[str, None] = None,
                            save_eps_name: Union[str, None] = None,
                            **kwargs) -> None:
    '''
    Plots the comparsion between spatial, nonspatial and RCTD weights and marker gene expression

    '''
    options = {
        'subplot_size': (3, 3),
        'save_eps_name': "subplots_layers_sseqcerebellum.eps",
        'dpi': 1200
    }
    options.update(kwargs)

    subplot_size = options['subplot_size']
    num_cols = len(cluster_keys)

    num_rows = 4
    if rctd_coord is None:
        num_rows -= 1
    if reference_dict is None:
        num_rows -= 1

    figsize_x = (subplot_size[0] * num_cols) * 1.1
    figsize_y = subplot_size[1] * num_rows

    print(f"\nGenerating {num_rows} by {num_cols} "
          f"grid of plots (figsize = {figsize_y}, {figsize_x})\n")

    fig = plt.figure(figsize=(figsize_x, figsize_y), constrained_layout=True)
    grid = fig.add_gridspec(ncols=num_cols, nrows=num_rows)

    for col, cluster_key in enumerate(cluster_keys):

        # BANKSY
        # ------

        ax = fig.add_subplot(grid[0, col])
        plot_cluster_subset(cluster_key[0], colours_dict, adata_spatial,
                            ax, spot_size=0.2, plot_umap=False)

        # non-spatial
        # -----------

        ax = fig.add_subplot(grid[1, col])
        plot_cluster_subset(cluster_key[1], colours_dict, adata_nonspatial,
                            ax, spot_size=0.2, plot_umap=False)

        # Plot scRNAseq metagene (from top DE genes)
        # ------------------------------------------

        if rctd_coord is not None and rctd_weights is not None and reference_dict is not None:

            ax = fig.add_subplot(grid[2, col])

            weights = np.zeros((len(rctd_coord),))

            for dv_clust in cluster_key[2]:
                clust_num = reference_dict[dv_clust] - 1
                weights += np.array(rctd_weights[:, clust_num]).flatten()

            _plot_continuous(rctd_coord.loc[:, "x"], rctd_coord["y"],
                             weights, ax, vmax=1, plot_cbar=False)

        # Plot scRNAseq metagene (from top DE genes)
        # ------------------------------------------

        if markergenes_dict is not None and adata_allgenes is not None:

            ax = fig.add_subplot(grid[-1, col])

            markers = []

            for sc_clust in cluster_key[-1]:

                # get a list of marker genes from the reference that are also in the dataset (unfiltered AnnData)
                markers_temp = [m for m in markergenes_dict[sc_clust] if m in adata_allgenes.var.index]

                if markers_temp < markergenes_dict[sc_clust]:
                    print(f"{len(markergenes_dict[sc_clust]) - len(markers_temp)} "
                          f"scRNA-seq DE genes in {sc_clust} absent/filtered from slideseq dataset")

                markers += markers_temp

            print(f"{len(markers)} scRNAseq DE markers for clusters {cluster_key[-1]}")

            # sum marker genes to obtain a metagene summarizing marker gene expression
            markers_slice = adata_allgenes[:, markers]
            metagene = np.mean(markers_slice.X, axis=1)

            _plot_continuous(adata_spatial.obs["xcoord"], adata_spatial.obs["ycoord"],
                             metagene, ax, vmax=1, plot_cbar=False)

    if save_fig:
        if save_png_name:
            fig.savefig(os.path.join(file_path, save_png_name), format='png', dpi=options['dpi'])
        if save_eps_name:
            fig.savefig(os.path.join(file_path, save_eps_name), format='eps', dpi=options['dpi'])


def compare_weights(adata_spatial: anndata.AnnData,
                    adata_nonspatial: anndata.AnnData,
                    metagene_df: anndata.AnnData,
                    rctd_coord: pd.DataFrame,
                    rctd_weights: pd.DataFrame,
                    cell_types: list,  # named cell-types derived from unsupervised (BANKSY/nonspatial) clustering
                    reference: list,  # names of cell-types in the reference data
                    cell_type_colname: str = "cell type refined",
                    reference_dict: dict = dropviz_dict,  # mapping of refernece cell-type names to rctd_weights column
                    save_fig: bool = False,
                    **kwargs) -> None:
    '''
    Plots a heatmap of correlations between
     (1) cell types assigned by unsupervised clsutering (BANKSY or non-spatial or other)
     (2) cell type composition estimates by reference-guided deconvolution (e.g. RCTD)
         or expression of marker gene(s) from a reference dataset
    Note that (1) is discrete and (2) is continous. Hence we measure the
    point-biserial correlation (equivalent to pearson correlation for continuous values)
    '''
    options = {
        'file_path': "data/slide_seq/v1",
        'dpi': 1200
    }
    options.update(kwargs)

    # join the dataframes to get the intersection of cells from RCTD and unsupervised clustering
    df_intersect = rctd_coord.join(adata_spatial.obs, how="inner")
    print(f"Number of cells in both RCTD and adata: {len(df_intersect)}")

    rctd_df = rctd_coord.copy()

    for cell_type in reference_dict:
        # column in weights matrix this cell type corresponds to
        col_num = reference_dict[cell_type] - 1
        rctd_df[cell_type] = rctd_weights[:, col_num]

    rctd_df = rctd_df.loc[df_intersect.index, :]

    metagene_df_intersect = metagene_df.loc[df_intersect.index, :]

    adata_spatial_rctd = adata_spatial[df_intersect.index, :]
    adata_nonspatial_rctd = adata_nonspatial[df_intersect.index, :]

    for comparison_df, s in [(rctd_df, "RCTD"), (metagene_df_intersect, "Metagene")]:

        print(f"Comparing to {s}")

        for adata, s2 in [(adata_nonspatial_rctd, "non-spatial"), (adata_spatial_rctd, "spatial")]:

            correlation_grid = np.zeros((len(cell_types), len(reference)), )
            pval_grid = np.zeros_like(correlation_grid)

            for col, ref in enumerate(reference):
                for row, compare in enumerate(cell_types):
                    mask = (adata.obs[cell_type_colname] == compare).astype(float).values

                    # compute point biserial correlation (r value and p-value)
                    r, pval = pointbiserialr(mask, comparison_df.loc[:, ref].values)
                    # print(f"Pearson correlation of {compare} to {ref} : r = {r}, p-value = {pval:0.3f}")

                    correlation_grid[row, col] = r
                    pval_grid[row, col] = pval

            fig, axs = plt.subplots(nrows=1, ncols=2,
                                    figsize=(len(reference) * 3.5 + 3, len(cell_types) * 0.9))

            # heatmap of r values
            heat_map = sns.heatmap(correlation_grid, linewidth=1,
                                   vmin=-.9, vmax=.9, cmap="bwr", annot=True, fmt="0.2f",
                                   yticklabels=cell_types, xticklabels=reference,
                                   square=True, ax=axs[0])

            # heatmap of p-values
            pvals = sns.heatmap(pval_grid, linewidth=1,
                                vmin=0, vmax=1,  # cmap = "bwr",
                                annot=True, fmt="0.1e",
                                yticklabels=cell_types, xticklabels=reference,
                                square=True, ax=axs[1])
            fig.suptitle(f"comparing {s2} clustering to {s}")
            fig.tight_layout()

            if save_fig:
                fig.savefig(os.path.join(options['file_path'], f"corr_to_{s}_{s2}.eps"),
                            format='eps', dpi=options['dpi'])


def plot_self_vs_nbr(adata_spatial: anndata.AnnData,
                     adata_self: anndata.AnnData,
                     metagene: anndata._core.views.ArrayView,
                     metagene_nbr: anndata._core.views.ArrayView,
                     metagene_nbr_1: anndata._core.views.ArrayView,
                     cell_type: str,
                     cell_type2: str,
                     mask_1: str,
                     mask_2: str,
                     lambda_param: int,
                     top_n: int = 20,
                     cell_type_colname: str = "cell type refined"
                     ) -> None:
    '''
    2D plot of metagene expression (self as x-axis and neighbour as y-axis)
    for 2 cell types using seabourn 'jointplot'
    '''

    dfs = []  # list of dataframes for each of the 2 clusters

    for mask, colour, cluster in [(mask_1, "red", cell_type), (mask_2, "blue", cell_type2)]:
        self_expr = np.sqrt(1 - lambda_param) * metagene[mask]
        nbr_expr = np.sqrt(lambda_param) * metagene_nbr[mask]
        nbr_expr_2 = np.sqrt(lambda_param) * metagene_nbr_1[mask]

        #     metagene_plot = ax.scatter(self_expr, nbr_expr, c=colour, s=.5, alpha=.2)
        dfs.append(
            pd.DataFrame(data={'self': self_expr, 'nbr': nbr_expr, "cluster": cluster})
        )

    df_combined = pd.concat(dfs)
    # df_combined["cluster"] = df_combined["cluster"].astype("category")
    # display(df_combined.dtypes)

    sns.jointplot(data=df_combined, x="self", y="nbr", hue="cluster", kind="kde",
                  joint_kws={"thresh": 0.03, "levels": 10, "bw_adjust": 1.5, "fill": False})


def plot_self_vs_nbr_metagene_all(
        adata_spatial: anndata.AnnData,
        adata_self: anndata.AnnData,
        reference_cluster,  # all clusters will be plotted against this one
        lambda_param: int,
        cell_type_colname: str = "cell type refined",
        max_m: int = 0,  # m=0 is neighbour average, m=1 is AGF
        top_n: int = 20,  # number of top genes to sum
        savefig: bool = False, ) -> None:
    '''
    2D plot of metagene expression (self as x-axis and neighbour as y-axis)
    for all cell types vs a reference cell type
    done with matplotlib, without using seabourn 'jointplot'
    '''

    if adata_self is None:
        adata_self = _adata_filter_self(adata_spatial, run_rank_genes_groups=True,
                                        cell_type_colname=cell_type_colname, )

    for m in range(max_m + 1):

        # Plot all clusters vs one cluster
        # --------------------------------

        cell_types = [cluster for cluster in adata_self.obs[cell_type_colname].cat.categories
                      if cluster != reference_cluster]

        print(f"Comparing <{reference_cluster}>  with:  {cell_types}\n")

        DE_genes1 = get_DEgenes(adata_self, reference_cluster, top_n=top_n)
        mask_1 = adata_spatial.obs[cell_type_colname] == reference_cluster

        subplot_size = (5, 5)
        num_cols = 2
        num_rows = len(cell_types) // num_cols + 1
        print(num_rows, num_cols)
        figsize_x = (subplot_size[0] * num_cols) * 1.1
        figsize_y = subplot_size[1] * num_rows

        fig = plt.figure(figsize=(figsize_x, figsize_y), constrained_layout=True)
        grid = fig.add_gridspec(ncols=num_cols, nrows=num_rows)

        for n, cell_type in enumerate(cell_types):

            subgrid = grid[n // num_cols, n % num_cols].subgridspec(5, 5)
            ax = fig.add_subplot(subgrid[1:, :-1])

            DE_genes2 = get_DEgenes(adata_self, cell_type, top_n=top_n, verbose=False)
            metagene, metagene_nbr = get_metagene_difference(
                adata_spatial, DE_genes1, DE_genes2, m=m
            )

            mask_2 = adata_spatial.obs[cell_type_colname] == cell_type

            dfs = []

            min_x, max_x = 0.0, 0.0

            for mask, colour, cluster in [(mask_1, "red", reference_cluster), (mask_2, "blue", cell_type), ]:
                self_expr = np.sqrt(1 - lambda_param) * metagene[mask]
                nbr_expr = np.sqrt(lambda_param) * metagene_nbr[mask]

                # plot scatterplot of all cells in 2D expression coordinates
                metagene_plot = ax.scatter(
                    self_expr, nbr_expr, c=colour, s=.6, alpha=0.1, label=cluster
                )

                dfs.append(
                    pd.DataFrame(data={'self': self_expr, 'nbr': nbr_expr, "cluster": cluster})
                )

                min_x = min(np.amin(self_expr), min_x)
                max_x = max(np.amax(self_expr), max_x)

            df_combined = pd.concat(dfs)

            # plot 2D density plot (this is the main plot)
            # --------------------------------------------

            sns.kdeplot(
                data=df_combined, x="self", y="nbr", hue="cluster", palette=["red", "blue"],
                alpha=0.5,  # common_norm = False, #common_grid=True,
                bw_adjust=1.2, thresh=0.05, levels=12,
                # multiple="stack",
                ax=ax,
            )

            # ax.set_xlim([min_x, max_x])
            # ax.set_ylim([min_x, max_x])
            ax.set_xlabel("self")
            ax.set_ylabel("neighbour")
            # ax.set_title(f"{cell_type_2} vs\n{cell_type}", fontsize=8)
            ax.spines['right'].set_visible(False)
            ax.spines['top'].set_visible(False)

            # plot 1D density plot on top of main plot
            # ----------------------------------------

            ax_top = fig.add_subplot(subgrid[:1, :-1], sharex=ax)
            sns.kdeplot(data=df_combined, x="self", hue="cluster", palette=["red", "blue"],
                        alpha=0.5, common_norm=False,  # linewidth = 0.2, #common_grid=True,
                        bw_adjust=1.2, legend=False,
                        # multiple="stack",
                        ax=ax_top)
            # ax_top.legend().set_visible(False)
            ax_top.axis("off")
            xmin, xmax = ax_top.get_xaxis().get_view_interval()
            ymin, _ = ax_top.get_yaxis().get_view_interval()
            ax_top.add_artist(
                plt.Line2D((xmin, xmax), (ymin, ymin), color='gray', linewidth=1, )
            )

            # plot 1D density plot on right of main plot
            # ------------------------------------------

            ax_right = fig.add_subplot(subgrid[1:, -1:], sharey=ax)
            sns.kdeplot(data=df_combined, y="nbr", hue="cluster", palette=["red", "blue"],
                        alpha=0.5, common_norm=False,  # linewidth = 0.2, #common_grid=True,
                        bw_adjust=1.2, legend=False,
                        # multiple="stack",
                        ax=ax_right)
            # ax_right.legend().set_visible(False)
            ax_right.axis("off")
            xmin, _ = ax_right.get_xaxis().get_view_interval()
            ymin, ymax = ax_right.get_yaxis().get_view_interval()
            ax_right.add_artist(
                plt.Line2D((xmin, xmin), (ymin, ymax), color='gray', linewidth=1, )
            )

        fig.tight_layout()

        if savefig:
            fig.savefig(f"metagene_nbr_plot_supp_m{m}.png", format='png', dpi=1200)



"""
Label Object
and functions dealing with labels
"""

import copy
import numpy as np

from typing import List, Tuple, Union

from scipy.sparse import csr_matrix, issparse
from scipy.optimize import linear_sum_assignment

import matplotlib as mpl
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.ticker import StrMethodFormatter
from matplotlib.collections import PolyCollection

from banksy_utils.time_utils import timer
from banksy.csr_operations import row_normalize


class Label(object):

    def __init__(self,
                 labels_dense: Union[np.ndarray, list],
                 verbose: bool = True,
                 ) -> None:

        self.verbose = verbose

        # Check type, dimensions, ensure all elements non-negative
        # --------------------------------------------------------

        if isinstance(labels_dense, list):
            labels_dense = np.asarray(labels_dense, dtype=np.int32)
        elif isinstance(labels_dense, np.ndarray):
            pass
        else:
            raise TypeError(
                f"Labels provided are of type {type(labels_dense)}. "
                f"Should be list or 1-dimensional numpy ndarray.\n"
            )

        assert labels_dense.ndim == 1, (
            f"Label array has {labels_dense.ndim} dimensions, "
            f"should be 1-dimensional."
        )
        assert np.issubdtype(labels_dense.dtype, np.integer), (
            f"Label array data type is {labels_dense.dtype}, "
            f"should be integer."
        )
        assert np.amin(labels_dense) >= 0, (
            f"Some of the labels have negative values.\n"
            f"All labels must be 0 or positive integers.\n"
        )

        # Initialize attributes
        # ---------------------

        # dense array of labels, same as provided
        self.dense = labels_dense
        # total number of data-points with label (e.g. number of cells)
        self.num_samples = len(labels_dense)

        # number of instances of each integer up to maximum label id
        self.bins = np.bincount(self.dense)
        # unique labels (all non-negative integers)
        self.ids = np.nonzero(self.bins)[0]
        # counts per label (same order as self.ids)
        self.counts = self.bins[self.ids]
        # highest integer id for a label
        self.max_id = np.amax(self.ids)
        # total number of labels
        self.num_labels = len(self.ids)

        self.onehot = None
        self.normalized_onehot = None

    def __repr__(self) -> str:
        return (
            f"{self.num_labels} labels, {self.num_samples} samples, "
            f"ids: {self.ids}, counts: {self.counts}"
        )

    def __str__(self) -> str:
        return (
            f"Label object:\n"
            f"Number of labels: {self.num_labels}, "
            f"number of samples: {self.num_samples}\n"
            f"ids: {self.ids}, counts: {self.counts},\n"
        )

    def get_onehot(self) -> csr_matrix:
        """
        return one-hot sparse array of labels.
        If not already computed, generate the sparse array from dense label array
        """
        if self.onehot is None:
            self.onehot = self.generate_onehot(verbose=False)

        return self.onehot

    def get_normalized_onehot(self) -> csr_matrix:
        """
        return normalized one-hot sparse array of labels.
        """
        if self.normalized_onehot is None:
            self.normalized_onehot = self.generate_normalized_onehot(
                verbose=False
            )

        return self.normalized_onehot

    def generate_normalized_onehot(self,
                                   verbose: bool = False,
                                   ) -> csr_matrix:
        """
        generate a normalized onehot matrix where each row is
        normalized by the count of that label
        e.g. a row [0 1 1 0 0] will be converted to [0 0.5 0.5 0 0]
        """
        return row_normalize(self.get_onehot().astype(np.float64),
                             copy=True,
                             verbose=verbose)

    def generate_onehot(self,
                        verbose: bool = False,
                        ) -> csr_matrix:
        """
        convert an array of labels to a
        num_labels x num_samples sparse one-hot matrix

        Labels MUST be integers starting from 0,
        but can have gaps in between e.g. [0,1,5,9]
        """

        # initialize the fields of the CSR
        # ---------------------------------

        indptr = np.zeros((self.num_labels + 1,), dtype=np.int32)
        indices = np.zeros((self.num_samples,), dtype=np.int32)
        data = np.ones_like(indices, dtype=np.int32)

        if verbose:
            print(f"\n--- {self.num_labels} labels, "
                  f"{self.num_samples} samples ---\n"
                  f"initalized {indptr.shape} index ptr: {indptr}\n"
                  f"initalized {indices.shape} indices: {indices}\n"
                  f"initalized {data.shape} data: {data}\n")

        # update index pointer and indices row by row
        # -------------------------------------------

        for n, label in enumerate(self.ids):

            label_indices = np.nonzero(self.dense == label)[0]
            label_count = len(label_indices)

            previous_ptr = indptr[n]
            current_ptr = previous_ptr + label_count
            indptr[n + 1] = current_ptr

            if verbose:
                print(f"indices for label {label}: {label_indices}\n"
                      f"previous pointer: {previous_ptr}, "
                      f"current pointer: {current_ptr}\n")

            if current_ptr > previous_ptr:
                indices[previous_ptr:current_ptr] = label_indices

        return csr_matrix((data, indices, indptr),
                          shape=(self.num_labels, self.num_samples))


#
#                  Relabeling Functions
# =====================================================
# Functions for swapping around or changing label ids
#
#

def _rand_binary_array(array_length, num_onbits):
    array = np.zeros(array_length, dtype=np.int32)
    array[:num_onbits] = 1
    np.random.shuffle(array)
    return array


@timer
def expand_labels(label: Label,
                  max_label_id: int,
                  sort_labels: bool = False,
                  verbose: bool = True,
                  ) -> Label:
    """
    Spread out label IDs such that they
    range evenly from 0 to max_label_id
    e.g. [0 1 2] -> [0 5 10]

    Useful if you need to be consistent with other
    label sets with many more label IDs.
    This spreads the out along the colour spectrum/map
    so that the colours are not too similar to each other.

    use sort_labels if the list of ids are not already sorted
    (which it should usually be)
    """
    if verbose:
        print(f"Expanding labels with ids: {label.ids} "
              f"so that ids range from 0 to {max_label_id}\n")

    if sort_labels:
        ids = np.sort(copy.copy(label.ids))
    else:
        ids = copy.copy(label.ids)

    # make sure smallest label-id is zero
    ids_zeroed = ids - np.amin(label.ids)

    num_extra_labels = max_label_id - np.amax(ids_zeroed)

    multiple, remainder = np.divmod(num_extra_labels, label.num_labels - 1)

    # insert regular spaces between each id
    inserted = np.arange(label.num_labels) * multiple

    # insert remaining spaces so that max label id equals given max_id
    extra = _rand_binary_array(label.num_labels - 1, remainder)

    expanded_ids = ids_zeroed + inserted
    expanded_ids[1:] += np.cumsum(extra)  # only add to 2nd label and above

    if verbose:
        print(f"Label ids zerod: {ids_zeroed}.\n"
              f"{multiple} to be inserted between each id: {inserted}\n"
              f"{remainder} extra rows to be randomly inserted: {extra}\n"
              f"New ids: {expanded_ids}")

    expanded_dense = (expanded_ids @ label.get_onehot()).astype(np.int32)

    return Label(expanded_dense, verbose=label.verbose)


@timer
def match_labels(labels_1: Label,
                 labels_2: Label,
                 extra_labels_assignment: str = "random",
                 verbose: bool = True,
                 ) -> Label:
    """
    Match second set of labels to first, returning a new Label object
    Uses scipy's version of the hungarian algroithm (linear_sum_assigment)
    """

    max_id = max(labels_1.max_id, labels_2.max_id)
    num_extra_labels = labels_2.num_labels - labels_1.num_labels

    if verbose:
        print(f"Matching {labels_2.num_labels} labels "
              f"against {labels_1.num_labels} labels.\n"
              f"highest label ID in both is {max_id}.\n")

    onehot_1, onehot_2 = labels_1.get_onehot(), labels_2.get_onehot()

    cost_matrix = (onehot_1 @ onehot_2.T).toarray()

    labels_match_1, labels_match_2 = linear_sum_assignment(cost_matrix,
                                                           maximize=True)

    if verbose:
        print("\nMatches:\n", list(zip(labels_match_1, labels_match_2)))

    # temporary list keeping track of which labels are still available for use
    available_labels = list(range(max_id + 1))
    # list to be filled with new label ids
    relabeled_ids = -1 * np.ones((labels_2.num_labels,), dtype=np.int32)

    # reassign labels
    # ---------------

    for index_1, index_2 in zip(labels_match_1, labels_match_2):
        label_1 = labels_1.ids[index_1]
        label_2 = labels_2.ids[index_2]
        if verbose:
            print(f"Assigning first set's {label_1} to "
                  f"second set's {label_2}.\n"
                  f"labels_left: {available_labels}")
        relabeled_ids[index_2] = label_1
        available_labels.remove(label_1)

    # Assign remaining labels (if 2nd has more labels than 1st)
    # ---------------------------------------------------------

    if num_extra_labels > 0:

        unmatched_indices = np.nonzero(relabeled_ids == -1)[0]

        assert num_extra_labels == len(unmatched_indices), (
            f"number of unmatched label IDs {len(unmatched_indices)} "
            f"does not match mumber of "
            f"extra labels in second set {num_extra_labels}.\n"
        )

        if extra_labels_assignment == "random":
            relabeled_ids[unmatched_indices] = np.random.choice(available_labels,
                                                                size=num_extra_labels,
                                                                replace=False, )

        elif extra_labels_assignment == "greedy":

            def _insert_label(array: np.ndarray,
                              max_length: int,
                              added_labels: list = [],
                              ) -> Tuple[np.ndarray, int, list]:
                """
                insert a label in the middle of the largest interval
                assumes array is alreay sorted!
                """
                if len(array) >= max_length:
                    return array, max_length, added_labels
                else:
                    intervals = array[1:] - array[:-1]
                    max_interval_index = np.argmax(intervals)
                    increment = intervals[max_interval_index] // 2
                    label_to_add = array[max_interval_index] + increment
                    inserted_array = np.insert(
                        array,
                        max_interval_index + 1,
                        label_to_add,
                    )
                    added_labels.append(label_to_add)
                    return _insert_label(inserted_array, max_length, added_labels)

            sorted_matched = np.sort(relabeled_ids[relabeled_ids != -1])

            if verbose:
                print(f"already matched ids (sorted): {sorted_matched}")

            _,_,added_labels = _insert_label(sorted_matched, labels_2.num_labels)

            relabeled_ids[unmatched_indices] = np.random.choice(added_labels,
                                                                size=num_extra_labels,
                                                                replace=False, )

        elif extra_labels_assignment == "optimized":
            raise NotImplementedError(
                f"haven't figured out how to do this yet.\n"
            )

        else:
            raise ValueError(
                f"Extra labels assignment method not recognised, "
                f"should be random, greedy or optimized.\n"
            )

        if verbose:
            print(f"\nRelabeled labels: {relabeled_ids}\n")

    # generate re-labeled dense array
    # -------------------------------
    #
    # reordered (matmul) sparse_labels -> dense relabelled array
    #   e.g.
    #   [ 1 3 2 ] @ [ 0 ... 1 0  -> [ 2 ... 1 3 ]
    #                 0 ... 0 1
    #                 1 ... 0 0]
    #

    relabeled_dense = (relabeled_ids @ onehot_2).astype(np.int32)

    return Label(relabeled_dense, verbose=labels_2.verbose)


def match_label_series(label_list: List[Label],
                       least_labels_first: bool = True,
                       extra_labels_assignment: str = "greedy",
                       verbose: bool = True,
                       ) -> Tuple[List[Label], int]:
    """
    Match a list of labels to each other, one after another
    in order of increasing (if least_labels_first is true)
    or decreasing (least_labels_first set to false)
    number of label ids.
    Returns the relabeled list in original order
    """
    num_label_list = [label.num_labels for label in label_list]
    max_num_labels = max(num_label_list)
    sort_indices = np.argsort(num_label_list)

    print(f"\nMaximum number of labels = {max_num_labels}\n"
          f"Indices of sorted list: {sort_indices}\n")

    ordered_relabels = []

    if least_labels_first:
        ordered_relabels.append(
            expand_labels(label_list[sort_indices[0]],
                          max_num_labels - 1)
        )
        if verbose:
            print(f"First label, expanded label ids: "
                  f"{ordered_relabels[0]}")
    else:
        # argsort is in asceding order, reverse it
        sort_indices = sort_indices[:, :, -1]
        # already has max number of labels, no need to expand
        ordered_relabels.append(
            label_list[sort_indices[0]]
        )

    for index in sort_indices[1:]:

        current_label = label_list[index]
        previous_label = ordered_relabels[-1]

        if verbose:
            print(f"\nRelabeling:\n{current_label}\n"
                  f"with reference to\n{previous_label}\n"
                  + "-" * 70 + "\n")

        relabeled = match_labels(
            previous_label, current_label,
            extra_labels_assignment=extra_labels_assignment,
            verbose=verbose,
        )

        ordered_relabels.append(relabeled)

    sort_indices_list = list(sort_indices)
    original_order_relabels = [ordered_relabels[sort_indices_list.index(n)]
                               for n in range(len(label_list))]

    return original_order_relabels, max_num_labels


#
#
#                           Label connections
# =========================================================================
#
# Functions to determine connections strengths between different labels
# Must be given:
#       (1) A Label object
#       (2) weights matrix representing the connections between each
#           sample (e.g. the spatial graph)
#
#

def interlabel_connections(label: Label,
                           weights_matrix: Union[csr_matrix, np.ndarray],
                           ) -> np.ndarray:
    """
    compute connections strengths between labels,
    normalized by number of each label
    Requires a weights_matrix of shape (num_samples x num_samples)
    reprensenting the spatial graph between each sample
    """
    assert weights_matrix.ndim == 2, (
        f"weights matrix has {weights_matrix.ndim} dimensions, should be 2."
    )

    assert weights_matrix.shape[0] == weights_matrix.shape[1] == label.num_samples, (
        f"weights matrix dimenisons do not match number of samples"
    )

    normalized_onehot = label.generate_normalized_onehot(verbose=False)

    print(
        f"\nmatrix multiplying labels x weights x labels-transpose "
        f"({normalized_onehot.shape} "
        f"x {weights_matrix.shape} "
        f"x {normalized_onehot.T.shape})\n"
    )

    connections = normalized_onehot @ weights_matrix @ normalized_onehot.T

    if issparse(connections):
        connections = connections.toarray()

    return connections


def plot_connections(label: Label,
                     weights_matrix: Union[csr_matrix, np.ndarray],
                     ax: mpl.axes.Axes,
                     zero_self_connections: bool = True,
                     normalize_by_self_connections: bool = False,
                     shapes_style: bool = True,
                     max_scale: float = 0.46,
                     colormap_name: str = "Spectral",
                     title_str="connection strengths between types",
                     title_fontsize: float = 12,
                     label_fontsize: float = 12,
                     verbose: bool = True,
                     ) -> None:
    """
    plot the connections between labels
    given as a num_label by num_label matrix of connection strengths

    :param ax: axes to plot on
    :param shapes_style: use shapes or heatmap
    :param max_scale: only used for shape, gives maximum size of square
    """

    fig = ax.get_figure()

    connections = interlabel_connections(label, weights_matrix)

    if zero_self_connections:
        np.fill_diagonal(connections, 0)
    elif normalize_by_self_connections:
        connections /= connections.diagonal()[:, np.newaxis]

    connections_max = np.amax(connections)

    # set label colours
    # -----------------

    cmap = mpl.cm.get_cmap(colormap_name)
    id_colours = {id: cmap(id / label.max_id) for id in label.ids}

    if shapes_style:

        # Coloured triangles with size representing connection strength
        # -------------------------------------------------------------

        left_triangle = np.array((
            (-1., 1.),
            # (1., 1.),
            (1., -1.),
            (-1., -1.)
        ))

        right_triangle = np.array((
            (-1., 1.),
            (1., 1.),
            (1., -1.),
            # (-1., -1.)
        ))

        polygon_list = []
        colour_list = []

        ax.set_ylim(- 0.55, label.num_labels - 0.45)
        ax.set_xlim(- 0.55, label.num_labels - 0.45)
        # ax.invert_yaxis()

        for label_1 in range(connections.shape[0]):
            for label_2 in range(connections.shape[1]):

                if label_1 <= label_2:

                    for triangle in [left_triangle, right_triangle]:
                        center = np.array((label_1, label_2))[np.newaxis, :]
                        scale_factor = connections[label_1, label_2] / connections_max
                        offsets = triangle * max_scale * scale_factor
                        polygon_list.append(center + offsets)

                    colour_list += (id_colours[label.ids[label_2]],
                                    id_colours[label.ids[label_1]])

        collection = PolyCollection(polygon_list,
                                    facecolors=colour_list,
                                    edgecolors="face",
                                    linewidths=0)

        ax.add_collection(collection)

        ax.tick_params(labelbottom=False, labeltop=True)
        ax.xaxis.set_tick_params(pad=-2)

    else:

        # Heatmap of connection strengths
        # -------------------------------

        heatmap = ax.imshow(connections,
                            cmap='viridis',
                            interpolation="nearest")

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.1)

        fig.colorbar(heatmap, cax=cax)
        cax.tick_params(axis='both', which='major', labelsize=6, rotation=-45)

        # change formatting if values too small
        if connections_max < 0.001:
            cax.yaxis.set_major_formatter(StrMethodFormatter('{x:,.1e}'))

    # Set tick labels positions and colour
    # ------------------------------------

    ax.set_aspect('equal')

    ax.set_xticks(np.arange(label.num_labels), )
    ax.set_xticklabels(label.ids, fontsize=label_fontsize,
                       fontweight="bold", rotation=0)

    ax.set_yticks(np.arange(label.num_labels), )
    ax.set_yticklabels(label.ids, fontsize=label_fontsize,
                       fontweight="bold")

    for ticklabels in [ax.get_xticklabels(), ax.get_yticklabels()]:
        for n, id in enumerate(label.ids):
            ticklabels[n].set_color(id_colours[id])

    # title
    # -----

    ax.set_title(title_str, fontsize=title_fontsize, fontweight="bold")


#
#                                        Test Script
# ==========================================================================================
#
#

if __name__ == "__main__":
    array1 = np.array([1, 1, 1, 1, 2, 2, 2, 3, 3, 3, 3])
    array2 = np.array([2, 2, 2, 1, 1, 1, 1, 5, 5, 5, 3])

    labels_1 = Label(array1, verbose=False)
    labels_2 = Label(array2, verbose=False)
    matched_labels_2 = match_labels(labels_1, labels_2, verbose=False)
    expanded = expand_labels(labels_1, 11)

    names = ["First", "second", "matched", "expanded first"]
    labels_list = [labels_1, labels_2, matched_labels_2, expanded]
    for name, label in zip(names, labels_list):
        print(f"\n {name}\n" + "-" * 50)
        print(label, "\n", label.dense,
              f"\nonehot:\n{label.get_onehot().toarray()}\n")

    print(f"\nOne-hot normalized:\n"
          f"{labels_2.get_normalized_onehot().toarray()}")

    weights_matrix = csr_matrix(np.random.random_sample((11, 11)))
    # print(weights_matrix)

    connections = interlabel_connections(expanded, weights_matrix)

    print(f"\nConnections Matrix:\n{connections}")

    import matplotlib.pyplot as plt

    fig, ax = plt.subplots()

    plot_connections(expanded, weights_matrix,
                     ax,
                     zero_self_connections=False)

    plt.show()



"""
BANKSY: Spatial Clustering Algorithm that Unifies Cell-Typing and Tissue Domain Segmentation

Building Aggregates with a Neighborhood Kernel and Spatial Yardstick (BANKSY) is a method
for clustering spatial transcriptomic data by augmenting the transcriptomic profile of each
cell with an average of the transcriptomes of its spatial neighbors.

For more details, see: https://github.com/prabhakarlab/Banksy_py
"""

__version__ = "1.3.3"
__author__ = "Nigel Chou, Yifei Yue, Vipul Singhal"

# Users should import from submodules:
# from banksy.initialize_banksy import initialize_banksy
# from banksy.run_banksy import run_banksy_multiparam
# from banksy.embed_banksy import generate_banksy_matrix

__all__ = ["__version__"]



import gc
import os
from typing import List, Union, Tuple

import anndata
from matplotlib import pyplot as plt, ticker as ticker

from banksy.embed_banksy import generate_banksy_matrix
from banksy.labels import Label
from banksy.main import concatenate_all
from banksy.cluster_methods import run_Leiden_partition, run_mclust_partition
from banksy_utils.umap_pca import pca_umap


def run_banksy_multiparam(adata: anndata.AnnData,
                          banksy_dict: dict,
                          lambda_list: List[int],
                          resolutions: List[int],
                          color_list: Union[List, str],
                          max_m: int,
                          filepath: str,
                          key: Tuple[str],
                          match_labels: bool = False,
                          pca_dims: List[int] = [20, ],
                          savefig: bool = True,
                          annotation_key: str = "cluster_name",
                          max_labels: int = None,
                          variance_balance: bool = False,
                          cluster_algorithm: str = 'leiden',
                          partition_seed: int = 1234,
                          add_nonspatial: bool = True,
                          **kwargs):
    options = {
        'save_all_h5ad': True,
        'save_name': 'slideseq_mousecerebellum_',
        'no_annotation': True,
        's': 50,
        'a': 1.0
    }

    options.update(kwargs)
    if not os.path.exists(filepath):
        os.makedirs(filepath)

    banksy_dict, banksy_matrix = generate_banksy_matrix(adata,
                                                        banksy_dict,
                                                        lambda_list,
                                                        max_m,
                                                        variance_balance=variance_balance)

    # Add nonspatial banksy matrix
    if add_nonspatial:
        banksy_dict["nonspatial"] = {0.0: {"adata": concatenate_all([adata.X], 0, adata=adata), }}

    pca_umap(banksy_dict,
             pca_dims=pca_dims,
             plt_remaining_var=False)

    if annotation_key:
        annotations = Label(adata.obs[annotation_key].cat.codes.tolist())
    else:
        annotations = None

    # Clustering algorithm
    if cluster_algorithm == 'leiden':
        print(f'Conducting clustering with Leiden Parition')
        results_df, max_num_labels = run_Leiden_partition(
            banksy_dict=banksy_dict,
            resolutions=resolutions,
            num_nn=50,
            num_iterations=-1,
            partition_seed=partition_seed,
            match_labels=match_labels,
            annotations=annotations,
            max_labels=max_labels
        )

    elif cluster_algorithm == 'mclust':
        print(f'Conducting clustering with mcluster algorithm')
        try:
            import rpy2
        except ModuleNotFoundError:
            print(f'Package rpy2 not installed, try pip install')

        match_labels = False
        results_df, max_num_labels = run_mclust_partition(
            banksy_dict=banksy_dict,
            partition_seed=partition_seed,
            match_labels=match_labels,
            annotations=annotations,
            num_labels=max_labels,
        )

    for params_name in results_df.index:
        gc.collect()

        adata_temp = results_df.loc[params_name, 'adata']
        raw_labels = results_df.loc[params_name, 'labels']
        num_clusters = results_df.loc[params_name, 'num_labels']
        lambda_p = results_df.loc[params_name, "lambda_param"]

        if annotation_key:
            ari_temp = results_df.loc[params_name, "ari"]
            ari_label = f'\nari = {round(ari_temp, 2)}'
        else:
            ari_label = ""

        raw_clusters = []

        if isinstance(raw_labels, Label):
            raw_labels = raw_labels.dense
        for i in raw_labels:
            raw_clusters.append(color_list[i])

        fig = plt.figure(figsize=(8, 8), constrained_layout=True)
        ax = fig.gca()
        title = f'λ = {round(lambda_p, 2)}' + ari_label

        print(f'Anndata {adata_temp.obsm}')
        subplot_sc(ax,
                   adata_temp,
                   key,
                   raw_clusters,
                   num_clusters,
                   title)

        if savefig:
            save_name = os.path.join(filepath, f'{params_name}.png')
            fig.savefig(save_name, format='png', dpi=108)

    return results_df


def subplot_sc(ax: plt.Axes,
               adata_temp: anndata.AnnData,
               coord_keys: Tuple[str],
               clusters_colormap: List,
               num_clusters: int,
               title: str) -> None:

    ax.scatter(adata_temp.obs[coord_keys[0]].values,
               adata_temp.obs[coord_keys[1]].values,
               c=clusters_colormap,
               vmin=0, vmax=num_clusters - 1,
               s=50000 / adata_temp.n_obs, alpha=1.0)

    ax.set_aspect('equal', 'datalim')
    ax.set_title(title, fontsize=20, fontweight="bold", )
    ax.axes.invert_yaxis()

    # Turn of ticks and frame
    ax.set(frame_on=False)
    ax.xaxis.set_major_locator(ticker.NullLocator())
    ax.yaxis.set_major_locator(ticker.NullLocator())



"""
Utility Function to load dataset as anndata object

Input: filepath, adata_filename

Returns: Annadata object

Yifei 15 May 2023
"""
# Use IPython display if available, otherwise fall back to print
try:
    from IPython.display import display
except ImportError:
    display = print

import os
import numpy as np
import pandas as pd
import scipy.sparse as sparse

from anndata import AnnData
import anndata
from typing import Tuple 

def load_adata(filepath: str,
               load_adata_directly: bool,
                adata_filename: str,
                gcm_filename: str = "", 
                locations_filename: str = "",
                coord_keys: Tuple[str] = ("xcoord", "ycoord", "xy_coord")
                ) -> anndata.AnnData:
    
    ### Add extra Option to load from adata or use existing Adata
    ### Add description to explain the function

    '''Utility Function to load dataset as anndata object,
        We assume that all files are in the same folder (filepath)

        Input Args: 
            - filepath (str): Path to folder of the file

            - load_adata_directly (bool): Whether to load annadata file directly, 
                If set to False, will attempt to convert gcm_file and locations_file as anndata object

            - adata_filename (str): File name of the designated Anndata object

        If current annadata is not present and we want to convert raw data to annadata
        
        Optional Args (For converting raw data to Anndata Object if it is not present):  
            - gcm_filename (str): .csv file containing the genome information of the cells
            - locations_filename (str): .csv file containing the x-y coordinates of the cells
            - coord_keys: A tuple of keys to index the x,y coordinates of the anndata object

        Returns: Loaded Annadata and the raw coordinates (raw_x, raw_y) as pd.Series'''
    if load_adata_directly:
            print(f"Looking for {os.path.join(filepath, adata_filename)}")
            if os.path.isfile(os.path.join(filepath, adata_filename)):
                
                # read existing anndata if present in folder
                # ------------------------------------------
                print("Attemping to read Annadata directly")
                adata = anndata.read_h5ad(os.path.join(filepath, adata_filename))
                print("Anndata file successfully loaded!")

            else:
                print(f"No such files {adata_filename} in {filepath}, please check the directory path and file names")
                print(f"Alternatively, try to convert raw files to anndata if by setting \'load_adata_directly = False\'")
    else:
        # If 'load_data_directly is set to false, try to read raw text files and convert to anndata
        try:
            gcm_df = pd.read_csv(os.path.join(filepath, gcm_filename), index_col = 0)
            locations_df = pd.read_csv(os.path.join(filepath, locations_filename), index_col = 0)
            print(f'GCM data successfully read as {gcm_df}\n Location data successfuly read as {locations_df}')
        except:
            Exception("Error occured when reading csv files, check the if files are permissible to read")

        
        sparse_X = sparse.csc_matrix(gcm_df.values.T)
        
        adata = AnnData(X = sparse_X, 
                        obs = locations_df, 
                        var = pd.DataFrame(index = gcm_df.index))
        
        adata.write(os.path.join(filepath, adata_filename))

        adata.obs[coord_keys[0]] = locations_df.loc[:, coord_keys[0]]
        adata.obs[coord_keys[1]] = locations_df.loc[:, coord_keys[1]]
        print("Anndata file successfully written!")
    
    ### Added keys,
    try:
        print(f'Attempting to concatenate spatial x-y under adata.obsm[{coord_keys[2]}]')
        x_coord, y_coord, xy_coord = coord_keys[0], coord_keys[1], coord_keys[2]
        raw_y, raw_x = adata.obs[y_coord], adata.obs[x_coord]
        adata.obsm[xy_coord] = np.vstack((adata.obs[x_coord].values,adata.obs[y_coord].values)).T
        print('Concatenation success!')
    except:
        print(f"Error in concatenating the matrices under adata.obsm[{coord_keys[2]}]\n raw_x, raw_y will return None")
        raw_y, raw_x = None, None

    return raw_y, raw_x, adata

def display_adata(adata: anndata.AnnData) -> None:
    '''
    Print summary / metadata of annadata object
    '''
    print("Displaying adata Object and their attributes")
    print("Adata attributes and dimensions:")
    display(adata)
    print(f"Matrix sparsity: {adata.X.nnz} filled elements "
      f"({adata.X.nnz/adata.X.shape[0]/adata.X.shape[1]:0.2f}) "
      f"out of {adata.X.shape[0]*adata.X.shape[1]}\n"
      f"max: {np.amax(adata.X.data)}, min: {np.amin(adata.X.data)}")

    # y and x locations / coordinates
    # -------------------------------
    print("\nDisplaying observations (adata.obs)")
    display(adata.obs)

    print("Displaying variables (adata.var)")
    display(adata.var)



"""
This script contains the function to carry out dimensionality reduction on the banksy matrices via PCA and UMAP

This has been modified to allow pca_umap to be run directly on single anndata objects instead of on the dictionary.
Use function pca_umap for dictionary format and pca_umap_adata for adata format.

needs to be tested further.

Yifei May 2023
modified by Nigel 7 Nov 2024
"""
from sklearn.decomposition import PCA
import umap
import anndata
import scipy.sparse as sparse
from scipy.sparse import csr_matrix, issparse
from banksy_utils.pca import plot_remaining_variance
from typing import List, Union
import numpy as np

def pca_umap(banksy_dict: dict,
             pca_dims: List[int] = [20,],
             plt_remaining_var: bool = True,
             add_umap: bool = False,
             **kwargs) -> None:
    '''
    Applies pca_umap to all adata objects in the banksy_dict

    Args:
        banksy_dict (dict): The processing dictionary containing info about the banksy matrices
    
    Optional Arg:
        pca_dims (List of integers): A list of integers which the PCA will reduce to
    
    Variable Args (kwargs):
        figsize (tuple of integers): A tuple for adjusting figure size

    Returns: Plot of remaining variance 
    '''
    
    options = {
        'figsize': (8,3)
    }
    print(f'Current decay types: {list(banksy_dict.keys())}')

    for nbr_weight_decay in banksy_dict:
        
        for lambda_param in banksy_dict[nbr_weight_decay]:
            
            if isinstance(lambda_param, str):
                continue # skip weights matrices

            print(
                f"\nReducing dims of dataset in (Index = {nbr_weight_decay}, lambda = {lambda_param})\n"
                + "=" * 50 + "\n"
            )
            
            # Retrieve anndata object
            # -----------------------
            
            adata_temp = banksy_dict[nbr_weight_decay][lambda_param]["adata"]
            
            param_str = f"(decay type = {nbr_weight_decay}, lambda = {lambda_param})"
            
            pca_umap_adata(adata_temp, pca_dims, plt_remaining_var, add_umap, options=options, param_str=param_str)

def pca_umap_adata(adata: anndata.AnnData,
                   pca_dims: List[int] = [20,],
                   plt_remaining_var: bool = True,
                   add_umap: bool = False,
                   options: dict = {'figsize': (8,3)},
                   param_str: str = "",
                   **kwargs) -> None:
    '''
    PCA_UMAP first applies dimensionality reduction via PCA,
    then applies UMAP to cluster the groups

    Args:
        banksy_dict (dict): The processing dictionary containing info about the banksy matrices
    
    Optional Arg:
        pca_dims (List of integers): A list of integers which the PCA will reduce to
    
    Variable Args (kwargs):
        figsize (tuple of integers): A tuple for adjusting figure size

    Returns: Plot of remaining variance 
    '''
    
    X = adata.X.todense() if issparse(adata.X) else adata.X
    
    X[np.isnan(X)] = 0

    # Reduce dimensions by PCA and then UMAP
    # --------------------------------------
                     
    if isinstance(pca_dims, int):
        pca_dims = [pca_dims,]
      
    for pca_dim in pca_dims:

        if isinstance(pca_dim, int):
            # We manually specify the number of PCs
            print(f"Setting the total number of PC = {pca_dim}")
            pca = PCA(n_components=pca_dim)
            reduced = pca.fit_transform(X)
        
        elif isinstance(pca_dim, float):
            # Otherwise, we specify the number of cumulative variance that each PC should represent
            print(f"Setting the total cumulative variance of PCs = {pca_dim}")
            pca = PCA(n_components=pca_dim)
            reduced = pca.fit_transform(X)

        #print(f'Noise Variance of PCA: {pca.noise_variance_}')
        #print(f'Variance of each PCs: {pca.explained_variance_ratio_}')
        print(
            f"Original shape of matrix: {X.shape}"
            f"\nReduced shape of matrix: {reduced.shape}\n" + "-" * 60 
            + f"\nmin_value = {reduced.min()}, mean = {reduced.mean()}, max = {reduced.max()}\n"
        )

        adata.obsm[f"reduced_pc_{pca_dim} {param_str}"] = reduced
        
        # Plot variance contribution for each component (elbow plot)
        # ----------------------------------------------------------
        if plt_remaining_var:
            plot_remaining_variance(
                pca, 
                figsize=options["figsize"], 
                title=f"remaining variance {param_str}"
            )

        # UMAP
        # ----

        if add_umap:
            print(f'Conducting UMAP and adding embeddings to adata.obsm["reduced_pc_{pca_dim}_umap"]')
            reducer = umap.UMAP(transform_seed = 42)
            umap_embedding = reducer.fit_transform(reduced)
            print(f"UMAP embedding\n" + "-" * 60  + f"\nshape: {umap_embedding.shape}\n\n")

            # save in AnnData object[obsm] attribute
            # ------------------
            adata.obsm[f"reduced_pc_{pca_dim}_umap {param_str}"] = umap_embedding                    
            print(adata.obsm)



"""Utility functions for BANKSY"""

# Users can import from submodules directly:
# from banksy_utils.load_data import load_adata
# from banksy_utils.filter_utils import filter_cells
# from banksy_utils.plot_utils import plot_cluster_results



"""
Utility Function to preprocess annadata 

Input: filepath, adata_filename

Returns: Annadata object

Yifei 15 May 2023
"""
from typing import Optional, Collection
from anndata import AnnData
import anndata
import scanpy as sc

def preprocess_data(adata: anndata.AnnData,
                    percent_top: Optional[Collection[int]] = (50, 100, 200, 500),
                    log1p: bool = True, 
                    inplace: bool = True):
    '''
    Preprocess Anndata object by removing duplicate names,
    creating a ["MT"] series for MT
    Also calculates the QC metrics and putting them in place in the adata object

    Input Args:
      adata: (Anndata object)

    Return: 
        pre-processed Annadata object
    '''
    adata.var_names_make_unique()
    adata.var["mt"] = adata.var_names.str.startswith("MT-")
    # Calulates QC metrics and put them in place to the adata object
    sc.pp.calculate_qc_metrics(adata, qc_vars=["mt"], 
                               percent_top=percent_top, 
                               log1p=log1p, 
                               inplace=inplace)
    return adata



import re, gc, time
import pandas as pd

# Use IPython display if available, otherwise fall back to print
try:
    from IPython.display import display
except ImportError:
    display = print

from banksy.main import LeidenPartition
from banksy.labels import Label, match_label_series
import anndata
import leidenalg
from sklearn.metrics.cluster import adjusted_rand_score

import sys
import warnings
try:
    import rpy2
except ModuleNotFoundError:
    warn_str = ("\nNo rpy2 installed. BANKSY will run, but mclust will not work.\n"+
                "Note: you can still use the default leiden option for clustering.\n"+
                "Install rpy2 and R in your conda environment if you want to use mclust for clustering.\n")
    warnings.warn(warn_str)

def run_mclust_partition(
        banksy_dict: dict,
        partition_seed: int = 1234,
        annotations = None,
        num_labels: int = None,
        **kwargs 
        ) -> dict: 
        '''
        Main driver function that runs the banksy computation; 
        1. Create neighbours graph in BANKSY Space
        2. Convert tos shared nearest neighbours
        3. Convert to igraph format for Leignalg package
        4. Leiden clustering is conducted using Leiden algorithm.

        Args:
            banksy_dict (dict): The processing dictionary containing:
            |__ nbr weight decay
            |__ lambda_param
                |__ anndata  

            resolutions: Resolution of the partition
                
            num_nn (int), default = 50: Number of nearest neighbours

            num_iterations (int), default = -1: 

            partition_seed (int): seed for partitioning (Leiden) algorithm
            
            match_labels (bool); default = True: Determines if labels are kept consistent across different hyperparameter settings

        Optional args (kwargs):
            Other parameters to the Leiden Partition

            shared_nn_max_rank (int), default = 3

            shared_nn_min_shared_nbrs (int), default = 5
        
        Returns:
            results_df (pd.DataFrame): A pandas dataframe containing the results of the partition
        '''

        if "rpy2" not in sys.modules:
            raise ModuleNotFoundError(
                "No rpy2 installed. Install rpy2 and R in your conda environment if you want to use mclust for clustering step")

        options = {
            'no_annotation': True,
        }
        options.update(kwargs)

        results = {}

        for nbr_weight_decay in banksy_dict:
            
            print(f"Decay type: {nbr_weight_decay}")

            for lambda_param in banksy_dict[nbr_weight_decay]:
                if not isinstance(lambda_param, float):
                    continue # skip other dictionary keys except lambda parameters

                print(f"Neighbourhood Contribution (Lambda Parameter): {lambda_param}")
            
                adata_temp = banksy_dict[nbr_weight_decay][lambda_param]["adata"]

                pca_dims = get_pca_dims(adata_temp)
                print("PCA dims to analyse:", pca_dims)

                for pca_dim in pca_dims:

                    if isinstance(pca_dim, str):
                        continue # skip full concatenated matrices
                    
                    print("\n" + "=" * 100 + 
                        f"\nSetting up partitioner for (nbr decay = {nbr_weight_decay}), "
                        f"Neighbourhood contribution = {lambda_param}, "
                        f"PCA dimensions = {pca_dim})\n" + "=" * 100 + "\n")
                    
                    used_obsm = f"reduced_pc_{pca_dim}"

                    print(f"\nFitting M-cluster algorithm with {num_labels} label\n" + "-" * 30 + "\n")

                    # Main partition via the m-cluster
                    # ---------------------------------------------------------------

                    print(f'Clustering PC_Embedding in obsm: {used_obsm}\n of type: {type(adata_temp.obsm[used_obsm])}')

                    adata = mclust_R(adata_temp, 
                                        num_labels, 
                                        model_names='EEE', 
                                        used_obsm=used_obsm, 
                                        random_seed=partition_seed)


                    # store results in dictionary
                    # ---------------------------------------------------------------

                    param_str = f"{nbr_weight_decay}_pc{pca_dim}_nc{lambda_param:0.2f}_labels{num_labels:0.2f}_mclust"
                    if (options['no_annotation']) and (not annotations):
                        print("No annotated labels")
                        results[param_str] = {
                            "decay": nbr_weight_decay,
                            "lambda_param": lambda_param,
                            "num_pcs":pca_dim,
                            "num_labels":num_labels,
                            "labels":adata.obs['mclust'],
                            "adata": banksy_dict[nbr_weight_decay][lambda_param]["adata"]
                        }

                    else:
                        print("Computing ARI for annotated labels")
                        results[param_str] = {
                            "decay":nbr_weight_decay,
                            "lambda_param": lambda_param,
                            "num_pcs": pca_dim,
                            "num_labels": num_labels,
                            "labels": adata.obs['mclust'],
                            "adata": banksy_dict[nbr_weight_decay][lambda_param]["adata"],
                            "ari" : adjusted_rand_score(annotations.dense, adata.obs['mclust']),
                        }
                
                        adata_temp.obs[param_str] = adata.obs['mclust']
                        print(f"Label Dense {adata.obs['mclust']}")
        
        results_df, max_num_labels = convert2df(results, match_labels = False,  resolution = False)
        return results_df, max_num_labels



def mclust_R(adata: anndata.AnnData,
              num_cluster: int,
              model_names: str ='EEE',
              used_obsm: str ='emb_pca',
              random_seed: int =1234):
        """\
        Clustering using the mclust algorithm.
        The parameters are the same as those in the R package mclust.

        This function is modified from STAGATE's original implementation of mclust
        see https://github.com/zhanglabtools/STAGATE/blob/main/STAGATE/utils.py
        """
        import numpy as np
        import rpy2.robjects as robjects
        import rpy2.robjects.numpy2ri as np2ri

        np.random.seed(random_seed)
        robjects.r.library("mclust")
        rpy2.robjects.numpy2ri.activate()
        r_random_seed = robjects.r['set.seed']
        r_random_seed(random_seed)
        rmclust = robjects.r['Mclust']
        
        embedding_np_array  = adata.obsm[used_obsm]
        print(f'obsm {used_obsm} : of type {type(embedding_np_array)} | shape {embedding_np_array.shape}')
        print(f'Array element is finite: {np.isfinite(embedding_np_array).all()}')

        r_embedding_vec = np2ri.numpy2rpy(embedding_np_array)

        res = rmclust(r_embedding_vec, num_cluster, model_names)
        mclust_res = np.array(res[-2]).astype('int')
        adata.obs['mclust'] = mclust_res.tolist()

        print(adata.obs['mclust'])
        return adata


def run_Leiden_partition(
        banksy_dict: dict,
        resolutions: list,          
        num_nn: int = 50,
        num_iterations: int = -1,
        partition_seed: int = 1234,
        match_labels: bool = True,
        annotations = None,
        max_labels: int = None,
        **kwargs) -> dict:
    '''
    Main driver function that runs Leiden partition across the banksy matrices stored in banksy_dict.
    See the original leiden package: https://leidenalg.readthedocs.io/en/stable/intro.html

    Args:
        banksy_dict (dict): The processing dictionary containing:
        |__ nbr weight decay
          |__ lambda_param
            |__ anndata  

        resolutions: Resolution of the partition
            
        num_nn (int), default = 50: Number of nearest neighbours

        num_iterations (int), default = -1: 

        partition_seed (int): seed for partitioning (Leiden) algorithm
        
        match_labels (bool); default = True: Determines if labels are kept consistent across different hyperparameter settings

    Optional args (kwargs):
        Other parameters to the Leiden Partition

        shared_nn_max_rank (int), default = 3

        shared_nn_min_shared_nbrs (int), default = 5
    
    Returns:
        results_df (pd.DataFrame): A pandas dataframe containing the results of the partition
    '''
    options = {
        'shared_nn_max_rank ': 3,
        'shared_nn_min_shared_nbrs' : 5,
        'verbose': True
    }
    options.update(kwargs)

    results = {}

    for nbr_weight_decay in banksy_dict:
        
        print(f"Decay type: {nbr_weight_decay}")

        for lambda_param in banksy_dict[nbr_weight_decay]:
            if not isinstance(lambda_param, float):
                continue # skip other dictionary keys except lambda parameters

            print(f"Neighbourhood Contribution (Lambda Parameter): {lambda_param}")
        
            adata_temp = banksy_dict[nbr_weight_decay][lambda_param]["adata"]

            pca_dims = get_pca_dims(adata_temp)
            print("PCA dims to analyse:", pca_dims)

            for pca_dim in pca_dims:
                
                if isinstance(pca_dim, str):
                    continue # skip full concatenated matrices
                
                print("\n" + "=" * 100 + 
                    f"\nSetting up partitioner for (nbr decay = {nbr_weight_decay}), "
                    f"Neighbourhood contribution = {lambda_param}, "
                    f"PCA dimensions = {pca_dim})\n" + "=" * 100 + "\n")
                
                banksy_reduced = adata_temp.obsm[f"reduced_pc_{pca_dim}"]
              
                partitioner = LeidenPartition(
                    banksy_reduced,
                    num_nn = num_nn,
                    nns_have_weights = True,
                    compute_shared_nn = True,
                    filter_shared_nn = True,
                    shared_nn_max_rank = options['shared_nn_max_rank '],
                    shared_nn_min_shared_nbrs = options['shared_nn_min_shared_nbrs'],
                    verbose = options['verbose'], 
                )

                if max_labels and not resolutions:
                    print("No resolutions indicated, trying to search for a suitable resolution from labels")
                    print(f"Finding resolution that matches {max_labels}")
                    resolutions = find_resolutions(
                        partitioner,
                        max_labels,
                        num_iterations,
                        partition_seed
                    )

                for resolution in resolutions:

                    print(f"\nResolution: {resolution}\n" + "-" * 30 + "\n")

                    # Main partition via the Leiden Algorithm
                    # ---------------------------------------------------------------

                    label, modularity = partitioner.partition(
                        resolution = resolution,
                        partition_metric = leidenalg.RBConfigurationVertexPartition,
                        n_iterations = num_iterations,
                        seed = partition_seed,
                    )

                    # store results in dictionary
                    # ---------------------------------------------------------------

                    param_str = f"{nbr_weight_decay}_pc{pca_dim}_nc{lambda_param:0.2f}_r{resolution:0.2f}"
                    if not annotations:
                        print("No annotated labels")
                        results[param_str] = {
                            "decay": nbr_weight_decay,
                            "lambda_param": lambda_param,
                            "num_pcs":pca_dim,
                            "resolution":resolution,
                            "num_labels": label.num_labels,
                            "labels": label,
                            "adata": banksy_dict[nbr_weight_decay][lambda_param]["adata"]
                        }

                    else:
                        print("Computing ARI for annotated labels")
                        results[param_str] = {
                            "decay": nbr_weight_decay,
                            "lambda_param": lambda_param,
                            "num_pcs": pca_dim,
                            "resolution": resolution,
                            "num_labels": label.num_labels,
                            "labels": label,
                            "adata": banksy_dict[nbr_weight_decay][lambda_param]["adata"],
                            "ari" : adjusted_rand_score(annotations.dense, label.dense),
                            # For determining ari of specific domains
                            #"ari_domain_straight" : adjusted_rand_score(adata_temp .obs["domain"].cat.codes.values, label.dense),
                            #"ari_domain_smooth" : adjusted_rand_score(adata_temp.obs["smoothed_manual"].cat.codes.values, label.dense),
                            #"ari_domain_smooth2" : adjusted_rand_score(adata_temp .obs["smoothed_manual2"].cat.codes.values, label.dense),
                        }
                
                        adata_temp.obs[param_str] = label.dense
                        
                        print(f"Label Dense {label.dense}")

    
    results_df, max_num_labels = convert2df(results, match_labels)

    return results_df, max_num_labels

def get_pca_dims(adata_temp):
    '''
    Auxiliary function to gets PCA dimensions from the anndata object
    '''
    pca_dims = []

    for key in adata_temp.obsm_keys():

        print(key, "\n")
        match = re.search(r"reduced_pc_([0-9]+)$", key)
        if match:
            pca_dims.append(int(match.group(1)))
            continue
        
        # else try to search for a float 
        match = re.search(r"reduced_pc_(0\.[0-9]+)$", key)
        if match:
            pca_dims.append(float(match.group(1)))
        
    return pca_dims


def convert2df(results: dict,
               match_labels: bool,
               resolution: bool = True) -> pd.DataFrame:
    '''
    Auxiliary function to convert results dictionary to pandas dataframe

    Also matchs labels if required
    '''
    results_df = pd.DataFrame.from_dict(results, orient = 'index')
    
    # sort by indices
    if resolution:
        sort_id = ["decay", "lambda_param", "num_pcs", "resolution"]
    else:
        sort_id = ["decay", "lambda_param", "num_pcs"]

    results_df.sort_values(by=sort_id, ascending=True, inplace=True)

    print("\nAfter sorting Dataframe")
    print(f"Shape of dataframe: {results_df.shape}")

    # find the maximum number of labels across all parameter sets
    # -----------------------------------------------------------
    ## If match labels are required
    if match_labels:
            results_df["relabeled"], max_num_labels = match_label_series(
            results_df["labels"],
            extra_labels_assignment="greedy",
            verbose=False
            )
            
            print('\nMatched Labels')
    
    max_num_labels = results_df["num_labels"].max()
    display(results_df)
    gc.collect()
    return results_df, max_num_labels

def find_resolutions(partitioner,
                    max_labels,
                    num_iterations,
                    partition_seed,
                    starting_resolution = 0.6,
                    step = 0.05,
                    max_iter = 25) -> int:
    
    '''Auxiliary function to find the resolution that matches the specified labels
    Returns a list containing a single resolution value that matches the number of clusters'''
    resolution = starting_resolution 
    label, _ = partitioner.partition(
                        resolution = resolution,
                        partition_metric = leidenalg.RBConfigurationVertexPartition,
                        n_iterations = num_iterations,
                        seed = partition_seed,
    )

    num_labels = label.num_labels
    # perform a simple linear start for labels
    iter = 1
    while (num_labels != max_labels) and (iter < max_iter):
        if num_labels < max_labels:
            # We have less labels than specified
            # increase resolution to increase the number of labels
            resolution += step
            label, _ = partitioner.partition(
                    resolution = resolution,
                    partition_metric = leidenalg.RBConfigurationVertexPartition,
                    n_iterations = num_iterations,
                    seed = partition_seed)
            num_labels = label.num_labels

        else:
            # There are now more labels then 
            resolution -= step
            label, _ = partitioner.partition(
                    resolution = resolution,
                    partition_metric = leidenalg.RBConfigurationVertexPartition,
                    n_iterations = num_iterations,
                    seed = partition_seed)
            num_labels = label.num_labels
        
        iter +=1
    
    print(f"Resolution found: {resolution} for labels: {max_labels}")
    return [resolution,]
